import { ipVersion, parseIp, stringifyIp, normalizeIp } from 'ip-bigint';

const bits = { 4: 32, 6: 128 };
function uniq(arr) {
  return Array.from(new Set(arr));
}
function cidrVersion(cidr) {
  return cidr.includes("/") ? ipVersion(cidr) : 0;
}
function compare(a, b) {
  const { number: aNum, version: aVersion } = parseIp(a.replace(/\/.+/, ""));
  const { number: bNum, version: bVersion } = parseIp(b.replace(/\/.+/, ""));
  if (aVersion === bVersion) {
    return aNum - bNum > 0n ? 1 : aNum - bNum < 0n ? -1 : 0;
  } else {
    return aVersion > bVersion ? 1 : 0;
  }
}
function doNormalize(cidr, { compress = true, hexify = false } = {}) {
  const { start, end, prefix, version } = parseCidr(cidr);
  if (start !== end) {
    const ip = normalizeIp(stringifyIp({ number: start, version }), { compress, hexify });
    return `${ip}/${prefix}`;
  } else {
    return normalizeIp(cidr, { compress, hexify });
  }
}
function normalizeCidr(cidr, opts) {
  if (Array.isArray(cidr)) {
    return cidr.map((entry) => normalizeCidr(entry, opts));
  } else {
    return doNormalize(cidr, opts);
  }
}
function parseCidr(str) {
  const cidrVer = cidrVersion(str);
  const parsed = /* @__PURE__ */ Object.create(null);
  let cidr;
  if (cidrVer) {
    cidr = str;
    parsed.version = cidrVer;
  } else {
    const version2 = ipVersion(str);
    if (version2) {
      cidr = `${str}/${bits[version2]}`;
      parsed.version = version2;
    } else {
      throw new Error(`Network is not a CIDR or IP: ${str}`);
    }
  }
  const [ipAndMisc, prefix] = cidr.split("/");
  if (!/^[0-9]+$/.test(prefix)) {
    throw new Error(`Network is not a CIDR or IP: ${str}`);
  }
  const { number, version, ipv4mapped, scopeid } = parseIp(ipAndMisc);
  parsed.ip = stringifyIp({ number, version, ipv4mapped, scopeid });
  parsed.cidr = `${parsed.ip}/${prefix}`;
  parsed.prefix = prefix;
  const numBits = bits[version];
  const ipBits = number.toString(2).padStart(numBits, "0");
  const prefixLen = Number(numBits - Number(prefix));
  const startBits = ipBits.substring(0, numBits - prefixLen);
  parsed.start = BigInt(`0b${startBits}${"0".repeat(prefixLen)}`);
  parsed.end = BigInt(`0b${startBits}${"1".repeat(prefixLen)}`);
  return parsed;
}
function doNetsOverlap(a, b) {
  if (a.start > b.end)
    return false;
  if (b.start > a.end)
    return false;
  return true;
}
function netContains(a, b) {
  if (b.start < a.start)
    return false;
  if (b.end > a.end)
    return false;
  return true;
}
function excludeNets(a, b, v) {
  const parts = [];
  if (a.start > b.end || a.end < b.start) {
    return [a.cidr];
  }
  if (a.start === b.start && a.end === b.end) {
    return [];
  }
  if (a.start > b.start && a.end < b.end) {
    return [];
  }
  if (a.start < b.start && a.end <= b.end) {
    parts.push({ start: a.start, end: b.start - 1n });
  }
  if (a.start >= b.start && a.end > b.end) {
    parts.push({ start: b.end + 1n, end: a.end });
  }
  if (a.start < b.start && a.end > b.end) {
    parts.push(
      { start: a.start, end: b.start - 1n },
      { start: b.end + 1n, end: a.end }
    );
  }
  const remaining = [];
  for (const part of parts) {
    for (const subpart of subparts(part)) {
      remaining.push(formatPart(subpart, v));
    }
  }
  return mergeCidr(remaining);
}
function biggestPowerOfTwo(num) {
  if (num === 0n)
    return 0n;
  return 2n ** BigInt(String(num.toString(2).length - 1));
}
function subparts(part) {
  if (part.end - part.start === 1n) {
    if (part.end % 2n === 0n) {
      return [{ start: part.start, end: part.start }, { start: part.end, end: part.end }];
    } else {
      return [{ start: part.start, end: part.end }];
    }
  }
  const size = diff(part.end, part.start);
  let biggest = biggestPowerOfTwo(size);
  let start;
  let end;
  if (size === biggest && part.start + size === part.end) {
    return [part];
  } else if (part.start % biggest === 0n) {
    start = part.start;
    end = start + biggest - 1n;
  } else {
    start = part.end / biggest * biggest;
    if (start + biggest - 1n > part.end) {
      start = (part.end / biggest - 1n) * biggest;
      while (start < part.start) {
        biggest /= 2n;
        start = (part.end / biggest - 1n) * biggest;
      }
      end = start + biggest - 1n;
    } else {
      start = part.end / biggest * biggest;
      end = start + biggest - 1n;
    }
  }
  let parts = [{ start, end }];
  if (start !== part.start) {
    parts = parts.concat(subparts({ start: part.start, end: start - 1n }));
  }
  if (end !== part.end) {
    parts = parts.concat(subparts({ start: end + 1n, end: part.end }));
  }
  return parts;
}
function diff(a, b) {
  a += 1n;
  return a - b;
}
function formatPart(part, version) {
  const ip = normalizeCidr(stringifyIp({ number: BigInt(part.start.toString()), version }));
  const zeroes = diff(part.end, part.start).toString(2);
  const prefix = bits[version] - (zeroes.match(/0/g) || []).length;
  return `${ip}/${prefix}`;
}
function mapNets(nets) {
  const maps = { 4: {}, 6: {} };
  for (const { start, end, version } of nets) {
    if (!maps[version][String(start)])
      maps[version][String(start)] = {};
    if (!maps[version][String(end)])
      maps[version][String(end)] = {};
    if (maps[version][String(start)].start) {
      maps[version][String(start)].start += 1;
    } else {
      maps[version][String(start)].start = 1;
    }
    if (maps[version][String(end)].end) {
      maps[version][String(end)].end += 1;
    } else {
      maps[version][String(end)].end = 1;
    }
  }
  return maps;
}
function doMerge(maps) {
  let start = null;
  let end = null;
  const numbers = Object.keys(maps);
  let depth = 0;
  const merged = [];
  for (const [index, number] of numbers.entries()) {
    const marker = maps[String(number)];
    if (start === null && marker.start) {
      start = BigInt(number);
    }
    if (marker.end) {
      end = BigInt(number);
    }
    if (marker.start)
      depth += marker.start;
    if (marker.end)
      depth -= marker.end;
    const next = numbers[index + 1];
    if (marker.end && depth === 0 && next && BigInt(next) - BigInt(number) > 1) {
      for (const sub of subparts({ start, end })) {
        merged.push(sub);
      }
      start = null;
      end = null;
    } else if (index === numbers.length - 1) {
      for (const sub of subparts({ start, end })) {
        merged.push(sub);
      }
    }
  }
  return merged;
}
function mergeCidr(nets) {
  const arr = uniq((Array.isArray(nets) ? nets : [nets]).sort(compare).map(parseCidr));
  const maps = mapNets(arr);
  const merged = { 4: [], 6: [] };
  for (const v of [4, 6]) {
    merged[v] = doMerge(maps[v]).map((part) => formatPart(part, v));
  }
  return [...merged[4].sort(compare), ...merged[6].sort(compare)];
}
function excludeCidr(base, excl) {
  const basenets = mergeCidr(uniq(Array.isArray(base) ? base : [base]));
  const exclnets = mergeCidr(uniq(Array.isArray(excl) ? excl : [excl]));
  const bases = { 4: [], 6: [] };
  const excls = { 4: [], 6: [] };
  for (const basenet of basenets) {
    const version = cidrVersion(basenet);
    if (version)
      bases[version].push(basenet);
  }
  for (const exclnet of exclnets) {
    const version = cidrVersion(exclnet);
    if (version)
      excls[version].push(exclnet);
  }
  for (const v of [4, 6]) {
    for (const exclcidr of excls[v]) {
      for (const [index, basecidr] of bases[v].entries()) {
        const base2 = parseCidr(basecidr);
        const excl2 = parseCidr(exclcidr);
        const remainders = excludeNets(base2, excl2, v);
        if (base2.cidr !== remainders.toString()) {
          bases[v] = bases[v].concat(remainders);
          bases[v].splice(index, 1);
        }
      }
    }
  }
  return bases[4].concat(bases[6]).sort(compare);
}
function expandCidr(nets) {
  const arr = uniq(Array.isArray(nets) ? nets : [nets]);
  const ips = [];
  for (const net of mergeCidr(arr)) {
    const { start, end, version } = parseCidr(net);
    for (let number = start; number <= end; number++) {
      ips.push(stringifyIp({ number, version }));
    }
  }
  return ips.map((ip) => normalizeCidr(ip));
}
function overlapCidr(a, b) {
  const aNets = uniq(Array.isArray(a) ? a : [a]);
  const bNets = uniq(Array.isArray(b) ? b : [b]);
  for (const a2 of aNets) {
    const aParsed = parseCidr(a2);
    for (const b2 of bNets) {
      const bParsed = parseCidr(b2);
      if (aParsed.version !== bParsed.version) {
        continue;
      }
      if (doNetsOverlap(aParsed, bParsed)) {
        return true;
      }
    }
  }
  return false;
}
function containsCidr(a, b) {
  const aNets = uniq(Array.isArray(a) ? a : [a]);
  const bNets = uniq(Array.isArray(b) ? b : [b]);
  const numExpected = bNets.length;
  let numFound = 0;
  for (const a2 of aNets) {
    const aParsed = parseCidr(a2);
    for (const b2 of bNets) {
      const bParsed = parseCidr(b2);
      if (aParsed.version !== bParsed.version) {
        continue;
      }
      if (netContains(aParsed, bParsed)) {
        numFound++;
        continue;
      }
    }
  }
  return numFound === numExpected;
}
const index = {
  mergeCidr,
  excludeCidr,
  expandCidr,
  overlapCidr,
  containsCidr,
  normalizeCidr,
  parseCidr
};

export { containsCidr, index as default, excludeCidr, expandCidr, mergeCidr, normalizeCidr, overlapCidr, parseCidr };
